"""Common fixtures for the aiobmsble library tests.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Awaitable, Buffer, Callable, Iterable
import logging
from types import ModuleType
from typing import Any, Final
from uuid import UUID

from _pytest.config import Notset
from bleak import BleakClient
from bleak.assigned_numbers import CharacteristicPropertyName
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
from bleak.exc import BleakCharacteristicNotFoundError
from bleak.uuids import normalize_uuid_str
from hypothesis import HealthCheck, settings
import pytest

from aiobmsble import BMSSample
from aiobmsble.utils import load_bms_plugins

logging.basicConfig(level=logging.INFO)
LOGGER: logging.Logger = logging.getLogger(__package__)

pytest_plugins: list[str] = ["aiobmsble.test_data"]


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add custom command-line option for max_examples."""
    parser.addoption(
        "--max-examples",
        action="store",
        type=int,
        default=1000,
        help="Set the maximum number of examples for Hypothesis tests.",
    )


def pytest_configure(config: pytest.Config) -> None:
    """Configure pytest with custom settings."""
    max_examples: int | Notset = config.getoption("--max-examples")
    settings.register_profile(
        "default",
        max_examples=max_examples,
        suppress_health_check=[HealthCheck.function_scoped_fixture],
    )
    settings.load_profile("default")


@pytest.fixture(
    params=sorted(
        load_bms_plugins(), key=lambda plugin: getattr(plugin, "__name__", "")
    ),
    ids=lambda param: param.__name__.rsplit(".", 1)[-1],
)
def plugin_fixture(request: pytest.FixtureRequest) -> ModuleType:
    """Return module of a BMS."""
    assert isinstance(request.param, ModuleType)
    return request.param


class MockBleakClient(BleakClient):
    """Mock bleak client."""

    BT_INFO: Final[dict[str, bytes]] = {
        "2a24": b"mock_model",
        "2a25": b"mock_serial_number",
        "2a26": b"mock_FW_version",
        "2a27": b"mock_HW_version",
        "2a28": b"mock_SW_version",
        "2a29": b"mock_manufacturer",
    }

    def __init__(
        self,
        address_or_ble_device: BLEDevice,
        disconnected_callback: Callable[[BleakClient], None] | None,
        services: Iterable[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Mock init."""
        LOGGER.debug("MockBleakClient init")
        super().__init__(
            address_or_ble_device.address
        )  # call with address to avoid backend resolving
        self._connected: bool = False
        self._notify_callback: Callable | None = None
        self._enabled_notify: set[BleakGATTCharacteristic | int | str | UUID] = set()
        self._disconnect_callback: Callable[[BleakClient], None] | None = (
            disconnected_callback
        )
        self._ble_device: BLEDevice = address_or_ble_device
        self._services: Iterable[str] | None = services

    @property
    def address(self) -> str:
        """Return device address."""
        return self._ble_device.address

    @property
    def is_connected(self) -> bool:
        """Mock connected."""
        return self._connected

    @property
    def services(self) -> BleakGATTServiceCollection:
        """Mock GATT services."""
        _services: BleakGATTServiceCollection = BleakGATTServiceCollection()
        _services.add_service(BleakGATTService(None, 0, normalize_uuid_str("180a")))
        return _services

    async def connect(self, **_kwargs: Any) -> None:
        """Mock connect."""
        assert not self._connected, "connect called, but client already connected."
        LOGGER.debug("MockBleakClient connecting %s", self._ble_device.address)
        self._connected = True

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
        ],
        **kwargs: Any,
    ) -> None:
        """Mock start_notify."""
        LOGGER.debug("MockBleakClient start_notify for %s", char_specifier)
        assert self._connected, "start_notify called, but client not connected."
        assert (
            char_specifier not in self._enabled_notify
        ), "start_notify called, but already notifying."
        self._notify_callback = callback
        self._enabled_notify.add(char_specifier)

    async def stop_notify(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID
    ) -> None:
        """Mock stop_notify."""
        LOGGER.debug("MockBleakClient stop_notify for %s", char_specifier)
        assert self._connected, "stop_notify called, but client not connected."
        self._notify_callback = None
        self._enabled_notify.remove(char_specifier)

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Mock write GATT characteristics."""
        LOGGER.debug(
            "MockBleakClient write_gatt_char %s, data: %s", char_specifier, data
        )
        assert self._connected, "write_gatt_char called, but client not connected."

    async def read_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        **kwargs: Any,
    ) -> bytearray:
        """Mock read GATT characteristics."""

        LOGGER.debug("MockBleakClient read_gatt_char %s", char_specifier)
        assert self._connected, "read_gatt_char called, but client not connected."

        if isinstance(char_specifier, str):
            char_specifier = normalize_uuid_str(char_specifier)[4:8]
            if char_specifier not in MockBleakClient.BT_INFO:
                raise BleakCharacteristicNotFoundError(char_specifier)
            return bytearray(MockBleakClient.BT_INFO[char_specifier])

        return bytearray()

    async def disconnect(self) -> None:
        """Mock disconnect."""

        LOGGER.debug("MockBleakClient disconnecting %s", self._ble_device.address)
        self._connected = False
        if self._disconnect_callback is not None:
            self._disconnect_callback(self)


@pytest.fixture(params=[-13, 0, 21], ids=["neg_current", "zero_current", "pos_current"])
def bms_data_fixture(request: pytest.FixtureRequest) -> BMSSample:
    """Return a fake BMS data dictionary."""

    return {
        "voltage": 7.0,
        "current": request.param,
        "cycle_charge": 34,
        "cell_voltages": [3.456, 3.567],
        "temp_values": [-273.15, 0.01, 35.555, 100.0],
    }


@pytest.fixture
def patch_bms_timeout(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[str | None, float], None]:
    """Fixture to patch BMS.TIMEOUT for different BMS classes."""

    def _patch_timeout(bms_class: str | None = None, timeout: float = 0.001) -> None:
        patch_class: str = (
            f"bms.{bms_class}.BMS.TIMEOUT"
            if bms_class
            else "basebms.BaseBMS._RETRY_TIMEOUT"
        )
        monkeypatch.setattr(f"aiobmsble.{patch_class}", timeout)

    return _patch_timeout


@pytest.fixture
def patch_bleak_client(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[type[BleakClient]], None]:
    """Fixture to patch BleakClient with a given MockClient."""

    def _patch(mock_client: type[BleakClient] = MockBleakClient) -> None:
        monkeypatch.setattr(
            "aiobmsble.basebms.BleakClient",
            mock_client,
        )

    return _patch


@pytest.fixture(params=[True, False], ids=["keep_alive", "reconnect"])
def keep_alive_fixture(request: pytest.FixtureRequest) -> bool:
    """Return True, False for keep_alive test."""
    assert isinstance(request.param, bool)
    return request.param


class DefGATTChar(BleakGATTCharacteristic):
    """Create BleakGATTCharacteristic with default values."""

    def __init__(
        self,
        handle: int,
        uuid: str,
        obj: Any = None,
        properties: list[CharacteristicPropertyName] | None = None,
        max_write_without_response_size: Callable[[], int] | None = None,
        service: BleakGATTService = BleakGATTService(
            None, 0, normalize_uuid_str("fff0")
        ),
    ) -> None:
        """Add default values for base class.

        Args:
            obj:
                A platform-specific object for this characteristic.
            handle:
                The handle for this characteristic.
            uuid:
                The UUID for this characteristic.
            properties:
                List of properties for this characteristic. Default: 'read'
            max_write_without_response_size:
                The maximum size in bytes that can be written to the characteristic
                in a single write without response command. Default: 512
            service:
                The service this characteristic belongs to. Default: 'fff0'

        """
        super().__init__(
            obj,
            handle,
            normalize_uuid_str(uuid),
            properties or ["read"],
            max_write_without_response_size or (lambda: 512),
            service,
        )



"""Test the E&J technology BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.basebms import crc_sum
from aiobmsble.bms.redodo_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

_RESULT_DEFS: Final[BMSSample] = {
    "voltage": 26.556,
    "current": -1.435,
    "cell_count": 8,
    "cell_voltages": [3.317, 3.319, 3.324, 3.323, 3.320, 3.314, 3.322, 3.317],
    "delta_voltage": 0.01,
    "power": -38.108,
    "battery_charging": False,
    "battery_level": 65,
    "battery_health": 100,
    "cycle_charge": 68.89,
    "design_capacity": 105,
    "cycle_capacity": 1829.443,
    "runtime": 172825,
    "temp_values": [23, 22, -2],
    "temp_sensors": 3,
    "temperature": 14.333,
    "cycles": 3,
    "problem": False,
    "problem_code": 0,
    "balancer": 0,
    "heater": False,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockRedodoBleakClient(MockBleakClient):
    """Emulate a Redodo BMS BleakClient."""

    _RESP = bytearray(
        b"\x00\x00\x65\x01\x93\x55\xaa\x00\x46\x66\x00\x00\xbc\x67\x00\x00\xf5\x0c\xf7\x0c\xfc\x0c"
        b"\xfb\x0c\xf8\x0c\xf2\x0c\xfa\x0c\xf5\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x65\xfa\xff\xff\x17\x00\x16\x00\xfe\xff\x00\x00\x00\x00\xe9\x1a\x04\x29"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x41\x00\x64\x00\x00\x00\x03\x00\x00\x00\x5f\x01\x00\x00\xa2"
    )  # 26.182V, cellVS 26.556V, 65%, 68.89Ah, 3 cycles

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffe2"):
            return bytearray()
        cmd: int = bytearray(data)[4]
        if cmd == 0x13:
            return self._RESP
        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        self._notify_callback(
            "MockRedodoBleakClient", self._response(char_specifier, data)
        )


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Redodo technology BMS data update."""

    patch_bleak_client(MockRedodoBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockRedodoBleakClient, BMS)


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            bytearray(
                b"\x00\x00\x65\x01\x93\x55\xaa\x00\x46\x66\x00\x00\xbc\x67\x00\x00\xf5\x0c\xf7\x0c"
                b"\xfc\x0c\xfb\x0c\xf8\x0c\xf2\x0c\xfa\x0c\xf5\x0c\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x65\xfa\xff\xff\x17\x00\x16\x00\x17\x00\x00\x00"
                b"\x00\x00\xe9\x1a\x04\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x41\x00\x64\x00\x00\x00\x03\x00\x00\x00"
                b"\x5f\x01\x00\x00\xff"
            ),
            "wrong CRC",
        ),
        (
            bytearray(
                b"\x00\x01\x65\x01\x93\x55\xaa\x00\x46\x66\x00\x00\xbc\x67\x00\x00\xf5\x0c\xf7\x0c"
                b"\xfc\x0c\xfb\x0c\xf8\x0c\xf2\x0c\xfa\x0c\xf5\x0c\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x65\xfa\xff\xff\x17\x00\x16\x00\x17\x00\x00\x00"
                b"\x00\x00\xe9\x1a\x04\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x41\x00\x64\x00\x00\x00\x03\x00\x00\x00"
                b"\x5f\x01\x00\x00\xbc"
            ),
            "wrong SOF",
        ),
        (
            bytearray(
                b"\x00\x00\x65\x01\x93\x55\xaa\x00\x46\x66\x00\x00\xbc\x67\x00\x00\xf5\x0c\xf7\x0c"
                b"\xfc\x0c\xfb\x0c\xf8\x0c\xf2\x0c\xfa\x0c\xf5\x0c\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x65\xfa\xff\xff\x17\x00\x16\x00\x17\x00\x00\x00"
                b"\x00\x00\xe9\x1a\x04\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x41\x00\x64\x00\x00\x00\x03\x00\x00\x00"
                b"\x5f\x01\x00\xbc"
            ),
            "wrong length",
        ),
        (bytearray(b"\x00"), "too short"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(MockRedodoBleakClient, "_RESP", wrong_response)

    patch_bleak_client(MockRedodoBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\x00\x00\x65\x01\x93\x55\xaa\x00\x46\x66\x00\x00\xbc\x67\x00\x00\xf5\x0c\xf7\x0c"
                b"\xfc\x0c\xfb\x0c\xf8\x0c\xf2\x0c\xfa\x0c\xf5\x0c\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x65\xfa\xff\xff\x17\x00\x16\x00\xfe\xff\x00\x00"
                b"\x00\x00\xe9\x1a\x04\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x41\x00\x64\x00\x00\x00\x03\x00\x00\x00"
                b"\x5f\x01\x00\x00\xa3"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\x00\x00\x65\x01\x93\x55\xaa\x00\x46\x66\x00\x00\xbc\x67\x00\x00\xf5\x0c\xf7\x0c"
                b"\xfc\x0c\xfb\x0c\xf8\x0c\xf2\x0c\xfa\x0c\xf5\x0c\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x65\xfa\xff\xff\x17\x00\x16\x00\xfe\xff\x00\x00"
                b"\x00\x00\xe9\x1a\x04\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x80\x00\x00\x00\x00\x00\x00\x41\x00\x64\x00\x00\x00\x03\x00\x00\x00"
                b"\x5f\x01\x00\x00\x22"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data up date with BMS returning protection flags."""

    monkeypatch.setattr(MockRedodoBleakClient, "_RESP", problem_response[0])

    patch_bleak_client(MockRedodoBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 63),
    }

    await bms.disconnect()


async def test_temp_sensor_detection(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test data up date with BMS returning protection flags."""

    _FRAME: bytearray = MockRedodoBleakClient._RESP.copy()

    patch_bleak_client(MockRedodoBleakClient)
    bms = BMS(generate_ble_device())

    for response, expected in (
        (b"\x00\x00\x00\x00", (1, [0])),  # all sensors zero
        (b"\x13\x00\x00\x00", (1, [19])),  # only first sensor non-zero
        (b"\x00\x00\x17\x00", (2, [0, 23])),  # only second sensor non-zero
        (b"\x17\x00", (2, [23, 0])),  # keep second sensor
        (b"\x15\x00".rjust(10, b"\x00"), (5, [0, 0, 0, 0, 21])),  # all sensors present
        (b"", (5, [0, 0, 0, 0, 0])),  # all sensors still present
    ):
        _FRAME[52:62] = response.ljust(10, b"\x00")  # set temp values
        _FRAME[-1] = crc_sum(_FRAME[:-1])  # update frame CRC
        monkeypatch.setattr(MockRedodoBleakClient, "_RESP", _FRAME)

        assert await bms.async_update() == _RESULT_DEFS | {
            "temp_sensors": expected[0],
            "temp_values": expected[1],
            "temperature": sum(expected[1]) / expected[0],
        }

    await bms.disconnect()



"""Package for battery management systems (BMS) via Bluetooth LE (aiobmsble).

Asynchronous Library to Query Battery Management Systems via Bluetooth LE

This library is intended to query data from battery management systems
that use Bluetooth LE. Stand-alone usage is possible in any Python environment
(with necessary dependencies installed).

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Callable
from contextlib import suppress
from enum import IntEnum
from importlib.metadata import PackageNotFoundError, version
from typing import Any, Literal, NamedTuple, TypedDict

__version__: str = "0.0.0.dev0"
with suppress(PackageNotFoundError):
    __version__ = version("aiobmsble")

type BMSValue = Literal[
    "battery_charging",
    "battery_mode",
    "battery_level",
    "battery_health",
    "current",
    "power",
    "temperature",
    "voltage",
    "cycles",
    "cycle_capacity",
    "cycle_charge",
    "total_charge",
    "delta_voltage",
    "problem",
    "runtime",
    "balancer",
    "balance_current",
    "cell_count",
    "cell_voltages",
    "design_capacity",
    "pack_count",
    "temp_sensors",
    "temp_values",
    "problem_code",
    "chrg_mosfet",
    "dischrg_mosfet",
    "heater",
]

type BMSpackvalue = Literal[
    "pack_voltages",
    "pack_currents",
    "pack_battery_levels",
    "pack_battery_health",
    "pack_cycles",
]


class BMSMode(IntEnum):
    """Enumeration of BMS modes."""

    UNKNOWN = -1
    BULK = 0x00
    ABSORPTION = 0x01
    FLOAT = 0x02


class BMSSample(TypedDict, total=False):
    """Dictionary representing a sample of battery management system (BMS) data."""

    battery_charging: bool  # True: battery charging
    battery_mode: BMSMode  # BMS charging mode
    battery_level: int | float  # [%] SoC
    battery_health: int | float  # [%] SoH
    current: float  # [A] (positive: charging)
    power: float  # [W] (positive: charging)
    temperature: int | float  # [°C]
    voltage: float  # [V]
    cycle_capacity: int | float  # [Wh]
    cycles: int  # [#]
    delta_voltage: float  # [V]
    problem: bool  # True: problem detected
    runtime: int  # [s]

    # detailed information
    balancer: bool | int  # False: off, True: active or bit mask, 1: enabled/active
    balance_current: float  # [A]
    cell_count: int  # [#] of parallel cells, i.e. per pack
    cell_voltages: list[float]  # [V]
    cycle_charge: int | float  # [Ah]
    total_charge: int  # [Ah], overall discharged
    design_capacity: int  # [Ah]
    pack_count: int  # [#]
    temp_sensors: int  # [#]
    temp_values: list[int | float]  # [°C]
    problem_code: int  # BMS specific code, 0 no problem, max. 64 bit

    # BMS switches
    chrg_mosfet: bool  # True: enabled
    dischrg_mosfet: bool  # True: enabled
    heater: bool  # True: enabled/heating

    # battery pack data
    pack_voltages: list[float]  # [V]
    pack_currents: list[float]  # [A]
    pack_battery_levels: list[int | float]  # [%]
    pack_battery_health: list[int | float]  # [%]
    pack_cycles: list[int]  # [#]


class BMSDp(NamedTuple):
    """Representation of one BMS data point."""

    key: BMSValue  # the key of the value to be parsed
    pos: int  # position within the message
    size: int  # size in bytes
    signed: bool  # signed value
    fct: Callable[[int], Any] = lambda x: x  # conversion function (default do nothing)
    idx: int = -1  # array index containing the message to be parsed


class BMSInfo(TypedDict, total=False):
    """Human readable information about the BMS device."""

    default_manufacturer: str
    default_model: str
    default_name: str
    fw_version: str
    manufacturer: str
    model: str
    model_id: str
    name: str
    serial_number: str
    sw_version: str
    hw_version: str


class MatcherPattern(TypedDict, total=False):
    """Optional patterns that can match Bleak advertisement data."""

    local_name: str  # name pattern that supports Unix shell-style wildcards
    manufacturer_data_start: list[int]  # start bytes of manufacturer data
    manufacturer_id: int  # required manufacturer ID
    oui: str  # required OUI used in the MAC address (first 3 bytes)
    service_data_uuid: str  # service data for the service UUID
    service_uuid: str  # 128-bit UUID that the device must advertise
    connectable: bool  # True if active connections to the device are required



"""Test the ANT implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.ant_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE: Final[int] = 20  # ANT BMS frame size

_RESULT_DEFS: Final[BMSSample] = {
    "cell_count": 22,
    "temp_sensors": 4,
    "voltage": 50.88,
    "current": 2.1,
    "battery_level": 10,
    "battery_health": 100,
    "cycle_charge": 9.957766,
    "design_capacity": 80,
    "total_charge": 15265,
    "cycles": 190,
    "temperature": 29.333,
    "cycle_capacity": 506.651,
    "power": 106.0,
    "battery_charging": False,
    "cell_voltages": [
        2.334,
        2.331,
        2.334,
        2.333,
        2.333,
        2.334,
        2.336,
        2.334,
        2.191,
        2.19,
        2.192,
        2.338,
        2.283,
        2.336,
        2.336,
        2.335,
        2.334,
        2.335,
        2.334,
        2.335,
        2.337,
        2.335,
    ],
    "temp_values": [29.0, 29.0, 29.0, 29.0, 30.0, 30.0],
    "delta_voltage": 0.148,
    "problem": False,
    "problem_code": 0,
    "balancer": False,
    "chrg_mosfet": True,
    "dischrg_mosfet": True,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockANTBleakClient(MockBleakClient):
    """Emulate a ANT BMS BleakClient."""

    CMDS: Final[dict[int, bytearray]] = {
        0x01: bytearray(b"\x7e\xa1\x01\x00\x00\xbe\x18\x55\xaa\x55"),
        0x02: bytearray(b"\x7e\xa1\x02\x6c\x02\x20\x58\xc4\xaa\x55"),
    }
    RESP: Final[dict[int, bytearray]] = {
        0x1: bytearray(
            b"\x7e\xa1\x11\x00\x00\x9e\x05\x04\x04\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x88\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1e\x09\x1b\x09\x1e\x09"
            b"\x1d\x09\x1d\x09\x1e\x09\x20\x09\x1e\x09\x8f\x08\x8e\x08\x90\x08\x22\x09\xeb\x08"
            b"\x20\x09\x20\x09\x1f\x09\x1e\x09\x1f\x09\x1e\x09\x1f\x09\x21\x09\x1f\x09\x1d\x00"
            b"\x1d\x00\x1d\x00\x1d\x00\x1e\x00\x1e\x00\xe0\x13\x15\x00\x0a\x00\x64\x00\x01\x01"
            b"\x00\x00\x00\xb4\xc4\x04\x86\xf1\x97\x00\x3a\xed\xe8\x00\x6a\x00\x00\x00\xa1\xf2"
            b"\xc0\x03\x00\x00\x00\x00\x22\x09\x0c\x00\x8e\x08\x0a\x00\x94\x00\x08\x09\x00\x00"
            b"\x6d\x00\x6a\x00\xaf\x02\xf3\xfa\x4c\x74\xe5\x00\x28\x66\xec\x00\xc3\x5d\x62\x00"
            b"\xcc\xb3\x92\x00\xed\xc8\xaa\x55"
        ),
        0x2: bytearray(
            b"\x7e\xa1\x12\x6c\x02\x20\x32\x34\x42\x48\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x32\x34\x42\x48\x55\x42\x30\x30\x2d\x32\x31\x31\x30\x32\x36\x41\x57\x96"
            b"\xff\x0b\x00\x00\x41\xf2\xaa\x55"
        ),
    }

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: Final[bytearray] = self.RESP.get(int(bytes(data)[2]), bytearray())
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockANTBleakClient", notify_data)


async def test_update(
    patch_bms_timeout, patch_bleak_client, keep_alive_fixture
) -> None:
    """Test ANT BMS data update."""

    patch_bms_timeout()
    patch_bleak_client(MockANTBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockANTBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "hw_version": "24BH",
        "sw_version": "24BHUB00-211026A",
    }


@pytest.fixture(
    name="wrong_response",
    params=[
        (b"\x6e" + MockANTBleakClient.RESP[0x1][1:], "wrong_SOF"),
        (b"\x7e\xa1\x12" + MockANTBleakClient.RESP[0x1][3:], "wrong_type"),
        (b"\x7e\xa1\x1f" + MockANTBleakClient.RESP[0x1][3:], "unknown_type"),
        (
            b"\x7e\xa1\x11\x00\x00\x01" + MockANTBleakClient.RESP[0x1][6:],
            "wrong_length",
        ),
        (MockANTBleakClient.RESP[0x1][:-2] + b"\xa1\x55", "wrong_EOF"),
        (b"\x7e\xa1\x11", "too_short"),
        (MockANTBleakClient.RESP[0x1][:-4] + b"\xff\xff\xaa\x55", "wrong_CRC"),
        (bytearray(1), "empty_response"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request):
    """Return faulty response frame."""
    return request.param[0]


async def test_invalid_response(
    monkeypatch, patch_bleak_client, patch_bms_timeout, wrong_response
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockANTBleakClient,
        "RESP",
        MockANTBleakClient.RESP | {0x1: wrong_response},
    )

    patch_bleak_client(MockANTBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\x7e\xa1\x11\x00\x00\x9e\x05\x04\x04\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x88\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1e\x09\x1b\x09\x1e\x09"
                b"\x1d\x09\x1d\x09\x1e\x09\x20\x09\x1e\x09\x8f\x08\x8e\x08\x90\x08\x22\x09\xeb\x08"
                b"\x20\x09\x20\x09\x1f\x09\x1e\x09\x1f\x09\x1e\x09\x1f\x09\x21\x09\x1f\x09\x1d\x00"
                b"\x1d\x00\x1d\x00\x1d\x00\x1e\x00\x1e\x00\xe0\x13\x15\x00\x0a\x00\x64\x00\x02\x02"
                b"\x00\x00\x00\xb4\xc4\x04\x86\xf1\x97\x00\x3a\xed\xe8\x00\x6a\x00\x00\x00\xa1\xf2"
                b"\xc0\x03\x00\x00\x00\x00\x22\x09\x0c\x00\x8e\x08\x0a\x00\x94\x00\x08\x09\x00\x00"
                b"\x6d\x00\x6a\x00\xaf\x02\xf3\xfa\x4c\x74\xe5\x00\x28\x66\xec\x00\xc3\x5d\x62\x00"
                b"\xcc\xb3\x92\x00\x4d\x04\xaa\x55"
            ),
            "low_value",
        ),
        (
            bytearray(
                b"\x7e\xa1\x11\x00\x00\x9e\x05\x04\x04\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x88\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1e\x09\x1b\x09\x1e\x09"
                b"\x1d\x09\x1d\x09\x1e\x09\x20\x09\x1e\x09\x8f\x08\x8e\x08\x90\x08\x22\x09\xeb\x08"
                b"\x20\x09\x20\x09\x1f\x09\x1e\x09\x1f\x09\x1e\x09\x1f\x09\x21\x09\x1f\x09\x1d\x00"
                b"\x1d\x00\x1d\x00\x1d\x00\x1e\x00\x1e\x00\xe0\x13\x15\x00\x0a\x00\x64\x00\x0e\x0e"
                b"\x00\x00\x00\xb4\xc4\x04\x86\xf1\x97\x00\x3a\xed\xe8\x00\x6a\x00\x00\x00\xa1\xf2"
                b"\xc0\x03\x00\x00\x00\x00\x22\x09\x0c\x00\x8e\x08\x0a\x00\x94\x00\x08\x09\x00\x00"
                b"\x6d\x00\x6a\x00\xaf\x02\xf3\xfa\x4c\x74\xe5\x00\x28\x66\xec\x00\xc3\x5d\x62\x00"
                b"\xcc\xb3\x92\x00\xc8\xf6\xaa\x55"
            ),
            "high_value",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bms_timeout,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockANTBleakClient,
        "RESP",
        MockANTBleakClient.RESP | {0x1: problem_response[0]},
    )

    patch_bleak_client(MockANTBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == _RESULT_DEFS | {
        "problem": True,
        "problem_code": (0x202 if problem_response[1] == "low_value" else 0xE0E),
        "chrg_mosfet": False,
        "dischrg_mosfet": False,
    }

    await bms.disconnect()



"""Test the aiobmsble library base class functions."""

from aiobmsble.test_data import bms_advertisements, ignore_advertisements
from aiobmsble.utils import bms_supported, load_bms_plugins


def test_advertisements_unique() -> None:
    """Check that each advertisement only matches one, the right BMS."""
    for adv, mac_addr, bms_real, _comments in bms_advertisements():
        for bms_under_test in load_bms_plugins():
            supported: bool = bms_supported(bms_under_test.BMS, adv, mac_addr)
            assert supported == (
                f"aiobmsble.bms.{bms_real}" == bms_under_test.__name__
            ), f"{adv} {"incorrectly matches"if supported else "does not match"} {bms_under_test}!"


def test_advertisements_ignore() -> None:
    """Check that each advertisement to be ignored is actually ignored."""
    for adv, mac_addr, reason, _comments in ignore_advertisements():
        for bms_under_test in load_bms_plugins():
            supported: bool = bms_supported(bms_under_test.BMS, adv, mac_addr)
            assert (
                not supported
            ), f"{adv} incorrectly matches {bms_under_test}! {reason=}"



"""Module to support Wattstunde Nova BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from string import hexdigits
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, b2str


class BMS(BaseBMS):
    """Wattstunde Nova Core BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Wattstunde",
        "default_model": "Nova Core",
    }
    _HEAD: Final[bytes] = b"\x3a"  # beginning of frame
    _TAIL: Final[bytes] = b"\x7e"  # end of frame
    _MIN_LEN: Final[int] = 238  # heater*2 + tail
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp(
            "current",
            10,
            4,
            False,
            lambda x: (x & 0x7FFF) / 1000 * (-1 if x >> 15 else 1),
        ),
        BMSDp("voltage", 18, 2, False, lambda x: x / 1000),
        BMSDp("cycles", 23, 2, False),
        BMSDp("battery_level", 25, 1, False),
        BMSDp("design_capacity", 26, 4, False, lambda x: x // 1000),
        BMSDp("cycle_charge", 30, 4, False, lambda x: x / 1000),
        BMSDp("heater", 115, 4, False, bool),
        BMSDp("problem_code", 0, 2, False, lambda x: x & 0x0FFC),
    )

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: bytes = b""

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "manufacturer_id": 28256,
                "manufacturer_data_start": [0x41],
                "connectable": True,
            }
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("FFF0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "FFF1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "FFF1"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        await self._await_msg(
            self._cmd(b"\x30\x31\x35\x31\x35\x30\x30\x30\x30\x45\x46\x45")
        )
        self._msg_event.clear()
        return BMSInfo(serial_number=b2str(self._msg[91:107]))

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if data.startswith(BMS._HEAD):
            self._frame.clear()

        self._frame += data

        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        if not (self._frame.startswith(BMS._HEAD) and self._frame.endswith(BMS._TAIL)):
            return

        if len(self._frame) % 2 or len(self._frame) < BMS._MIN_LEN:
            self._log.debug("incorrect frame length (%i)", len(self._frame))
            return

        if not all(chr(c) in hexdigits for c in self._frame[1:-1]):
            self._log.debug("incorrect frame encoding.")
            self._frame.clear()
            return

        _key: Final[int] = int(self._frame[7:9], 16)
        _dec = bytes(b ^ _key for b in bytes.fromhex(self._frame[1:-3].decode("ascii")))
        # incoming frames seem to have invalid checksum, thus not checked here

        if not _dec.startswith(b"\x01\x54"):
            self._log.debug("incorrect frame type.")
            self._frame.clear()
            return

        self._msg = _dec
        self._msg_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: bytes) -> bytes:
        """Assemble a Wattstunde Nova BMS command frame."""
        return BMS._HEAD + cmd + BMS._TAIL

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        if not self._msg_event.is_set():
            self._log.debug("requesting BMS data")
            await self._await_msg(
                self._cmd(b"\x30\x31\x35\x31\x35\x30\x30\x30\x30\x45\x46\x45")
            )

        result: BMSSample = BMS._decode_data(BMS._FIELDS, self._msg, start=44)
        result["cell_voltages"] = BMS._cell_voltages(self._msg, cells=16, start=12)
        result["temp_values"] = BMS._temp_values(
            self._msg, values=4, start=48, size=1, signed=False, offset=40
        )
        self._msg_event.clear()
        return result



"""Test the Gobel Power BLE BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSInfo, BMSSample
from aiobmsble.basebms import crc_modbus
from aiobmsble.bms.gobel_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

# Service and characteristic UUIDs for Gobel Power BMS
SERVICE_UUID = "00002760-08c2-11e1-9073-0e8ac72e1001"
TX_CHAR_UUID = "00002760-08c2-11e1-9073-0e8ac72e0001"
RX_CHAR_UUID = "00002760-08c2-11e1-9073-0e8ac72e0002"

BT_FRAME_SIZE = 20

_FRAME_MAIN_DATA: Final[bytearray] = bytearray(
    b"\x01\x03\x76\x00\x00\x05\x33\x00\x61\x00\x64\x7b\x11\x7e\xed\x7a"
    b"\xa8\x00\x01\x00\x00\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\xc0"
    b"\x00\x00\x04\x0c\xff\x0d\x02\x0d\x02\x0d\x01\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x01\x00\xd1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\xd3\xff\xff\x01\xd6"
)

_FRAME_DEVICE_INFO: Final[bytearray] = bytearray(
    b"\x01\x03\x46\x50\x34\x53\x32\x30\x30\x41\x2d\x34\x30\x35\x36\x39"
    b"\x2d\x31\x2e\x30\x32\x00\x00\x34\x30\x35\x36\x39\x31\x31\x41\x31"
    b"\x31\x30\x30\x30\x33\x32\x50\x20\x20\x20\x20\x47\x50\x2d\x4c\x41"
    b"\x31\x32\x2d\x33\x31\x34\x32\x30\x32\x35\x30\x36\x31\x38\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\xa3"
)

# Expected result from parsing _FRAME_MAIN_DATA
_RESULT_MAIN_DATA: Final[BMSSample] = {
    "battery_charging": False,
    "battery_health": 100,
    "battery_level": 97,
    "cell_count": 4,
    "cell_voltages": [3.327, 3.33, 3.33, 3.329],
    "chrg_mosfet": True,
    "current": 0.0,
    "cycle_capacity": 4193.316,
    "cycle_charge": 315.05,
    "cycles": 1,
    "delta_voltage": 0.003,
    "design_capacity": 324,
    "dischrg_mosfet": True,
    "power": 0.0,
    "problem": False,
    "problem_code": 0,
    "temp_sensors": 1,
    "temp_values": [20.9, 21.1],
    "temperature": 21.0,
    "voltage": 13.31,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockGobelBleakClient(MockBleakClient):
    """Emulate a Gobel Power BMS BleakClient."""

    _RESP: bytearray = _FRAME_MAIN_DATA
    _RESP_DEVICE_INFO: bytearray = _FRAME_DEVICE_INFO

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        """Generate response based on command."""
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str(TX_CHAR_UUID):
            req_data = bytes(data)
            # Verify it's a valid Modbus read request
            if len(req_data) >= 8 and req_data[0] == 0x01 and req_data[1] == 0x03:
                # Check which command is being sent
                start_addr = (req_data[2] << 8) | req_data[3]
                if start_addr == 0x00AA:  # Device info command (READ_CMD_3)
                    return bytearray(self._RESP_DEVICE_INFO)
                return bytearray(self._RESP)
        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        self._notify_callback(
            "MockGobelBleakClient", self._response(char_specifier, data)
        )


class MockFragmentedBleakClient(MockGobelBleakClient):
    """Emulate a Gobel Power BMS that sends fragmented responses."""

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT and send fragmented response."""
        await MockBleakClient.write_gatt_char(self, char_specifier, data, response)
        assert self._notify_callback is not None
        full_response = self._response(char_specifier, data)
        # Send in BT_FRAME_SIZE-byte chunks to simulate BLE fragmentation
        for i in range(0, len(full_response), BT_FRAME_SIZE):
            chunk = full_response[i : i + BT_FRAME_SIZE]
            self._notify_callback("MockFragmentedBleakClient", chunk)


async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    keep_alive_fixture: bool,
) -> None:
    """Test Gobel Power BMS data update."""

    monkeypatch.setattr(MockGobelBleakClient, "_RESP", _FRAME_MAIN_DATA)
    patch_bleak_client(MockGobelBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_MAIN_DATA

    # Query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_fragmented_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test handling of fragmented BLE responses."""

    monkeypatch.setattr(MockFragmentedBleakClient, "_RESP", _FRAME_MAIN_DATA)
    patch_bleak_client(MockFragmentedBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_MAIN_DATA

    await bms.disconnect()


async def test_device_info(monkeypatch: pytest.MonkeyPatch, patch_bleak_client) -> None:
    """Test fetching device info from BMS via Modbus."""

    monkeypatch.setattr(MockGobelBleakClient, "_RESP_DEVICE_INFO", _FRAME_DEVICE_INFO)
    patch_bleak_client(MockGobelBleakClient)

    bms = BMS(generate_ble_device())
    info: Final[BMSInfo] = await bms.device_info()

    assert info.get("sw_version") == "P4S200A-40569-1.02"
    assert info.get("serial_number") == "4056911A1100032P"
    assert info.get("model_id") == "GP-LA12-31420250618"

    await bms.disconnect()


def _corrupt_crc(frame: bytearray) -> bytearray:
    """Corrupt the CRC of a Modbus frame."""
    result = bytearray(frame)
    result[-1] ^= 0xFF
    return result


def _modbus_error_response() -> bytearray:
    """Build a Modbus error response."""
    error_frame = bytearray([0x01, 0x83, 0x02])  # Illegal data address
    error_frame.extend(crc_modbus(error_frame).to_bytes(2, "little"))
    return error_frame


def _short_response() -> bytearray:
    """Build a valid but short response (insufficient data)."""
    short_data = bytearray(20)
    short_data[2:4] = (1332).to_bytes(2, "big")  # voltage
    frame: bytearray = bytearray([0x01, 0x03, len(short_data)]) + short_data
    frame.extend(crc_modbus(frame).to_bytes(2, "little"))
    return frame


def _wrong_function_code() -> bytearray:
    """Build response with wrong function code."""
    return bytearray([0x01, 0x10, 0x00, 0x00, 0x00, 0x3B])


@pytest.mark.parametrize(
    ("wrong_response", "expect_timeout"),
    [
        (b"", True),
        (_corrupt_crc(_FRAME_MAIN_DATA), True),
        (_modbus_error_response(), True),
        (_wrong_function_code(), True),
        (_short_response(), False),  # Returns empty result, no timeout
    ],
    ids=[
        "empty",
        "invalid_crc",
        "modbus_error",
        "wrong_func_code",
        "short_response",
    ],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
    expect_timeout: bool,
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(MockGobelBleakClient, "_RESP", bytearray(wrong_response))
    patch_bleak_client(MockGobelBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    if expect_timeout:
        with pytest.raises(TimeoutError):
            result = await bms.async_update()
    else:
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


def test_bms_info() -> None:
    """Test BMS info definition."""
    assert BMS.INFO.get("default_manufacturer") == "Gobel Power"
    assert BMS.INFO.get("default_model") == "BLE BMS"


def test_cmd() -> None:
    """Test Modbus read command building."""
    assert BMS._cmd(0x01, 0x03, 0x0000, 0x003B) == b"\x01\x03\x00\x00\x00\x3b\x04\x19"


def _build_test_frame(reg_values: dict[int, int]) -> bytearray:
    """Build a test frame with specific register values for edge case testing.

    Args:
        reg_values: Dict mapping register offset to 16-bit value

    Returns:
        Complete Modbus frame with CRC.

    """
    data = bytearray(118)  # 59 registers = 118 bytes
    # Set default valid values
    data[2:4] = (1332).to_bytes(2, "big")  # voltage
    data[4:6] = (97).to_bytes(2, "big")  # SOC
    data[28:30] = (0xC000).to_bytes(2, "big")  # MOS status (both on)

    # Apply custom register values
    for reg, val in reg_values.items():
        data[reg * 2 : reg * 2 + 2] = val.to_bytes(2, "big")

    frame: bytearray = bytearray([0x01, 0x03, len(data)]) + data
    frame.extend(crc_modbus(frame).to_bytes(2, "little"))
    return frame


@pytest.mark.parametrize(
    ("reg_values", "expected_key", "expected_value"),
    [
        # Single cell - delta should be 0
        ({15: 1, 16: 3329}, "delta_voltage", 0),
        # Zero cells - empty cell_voltages list
        ({15: 0}, "cell_voltages", []),
        # Zero temp sensors with invalid MOSFET temp - empty temp_values list
        ({46: 0, 57: 0xFFFF}, "temp_values", []),
        # Alarm flags set (6-byte big-endian: 0x000100020003)
        ({8: 0x0001, 9: 0x0002, 10: 0x0003}, "problem_code", 0x000100020003),
        # Zero capacity values are included as 0 (framework pattern)
        ({4: 0, 5: 0}, "cycle_charge", 0.0),
    ],
    ids=[
        "single_cell_delta_zero",
        "no_cells",
        "no_temps_invalid_mos",
        "alarm_flags",
        "zero_capacity",
    ],
)
async def test_edge_cases(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    reg_values: dict[int, int],
    expected_key: str,
    expected_value,
) -> None:
    """Test various edge cases with parametrized register values."""

    frame: bytearray = _build_test_frame(reg_values)
    monkeypatch.setattr(MockGobelBleakClient, "_RESP", frame)
    patch_bleak_client(MockGobelBleakClient)

    bms = BMS(generate_ble_device())
    result: BMSSample = await bms.async_update()

    assert result.get(expected_key) == expected_value

    await bms.disconnect()


async def test_mos_temp_ffff_invalid(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test that MOSFET temperature of 0xFFFF is treated as invalid."""

    # Frame with 1 temp sensor at 25.0°C and invalid MOSFET temp
    frame: bytearray = _build_test_frame({46: 1, 47: 250, 57: 0xFFFF})
    monkeypatch.setattr(MockGobelBleakClient, "_RESP", frame)
    patch_bleak_client(MockGobelBleakClient)

    bms = BMS(generate_ble_device())
    result: BMSSample = await bms.async_update()

    # Should only have one temperature (not the invalid MOSFET temp)
    assert result.get("temp_values", []) == [25.0]

    await bms.disconnect()


def _build_device_info(length: int) -> bytearray:
    """Build device info frame with less than 60 bytes of data."""
    frame: bytearray = bytearray([0x01, 0x03, length]) + bytes(length)
    frame.extend(crc_modbus(frame).to_bytes(2, "little"))
    return frame


@pytest.mark.parametrize(
    "device_info_frame",
    [
        _build_device_info(40),
        _build_device_info(70),
        bytearray(),  # No response (timeout)
    ],
    ids=[
        "short_data",
        "empty_strings",
        "timeout",
    ],
)
async def test_device_info_edge_cases(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    device_info_frame: bytearray,
) -> None:
    """Test device info handling with various edge cases."""

    patch_bms_timeout()
    monkeypatch.setattr(MockGobelBleakClient, "_RESP_DEVICE_INFO", device_info_frame)
    patch_bleak_client(MockGobelBleakClient)

    bms = BMS(generate_ble_device())
    info: Final[BMSInfo] = await bms.device_info()

    # For short/timeout cases, model_id should not be present
    # For empty_strings case, model_id is present but empty
    if len(device_info_frame) >= 65:
        assert info.get("model_id") == ""
    else:
        assert "model_id" not in info

    await bms.disconnect()


async def test_unexpected_data_ignored(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test that unexpected data without a started frame is ignored."""

    class MockUnexpectedDataBleakClient(MockGobelBleakClient):
        """Send unexpected data before valid response."""

        async def write_gatt_char(
            self,
            char_specifier: BleakGATTCharacteristic | int | str | UUID,
            data: Buffer,
            response: bool | None = None,
        ) -> None:
            """Issue write command and send unexpected continuation data."""
            await MockBleakClient.write_gatt_char(self, char_specifier, data, response)
            assert self._notify_callback is not None
            # Send data that looks like a continuation (doesn't start with slave addr)
            self._notify_callback(
                "MockUnexpectedDataBleakClient", bytearray([0xFF, 0xFF])
            )
            # Then send the valid response
            self._notify_callback(
                "MockUnexpectedDataBleakClient", bytearray(self._RESP)
            )

    monkeypatch.setattr(MockUnexpectedDataBleakClient, "_RESP", _FRAME_MAIN_DATA)
    patch_bleak_client(MockUnexpectedDataBleakClient)

    bms = BMS(generate_ble_device())
    assert await bms.async_update() == _RESULT_MAIN_DATA

    await bms.disconnect()


async def test_short_initial_frame(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test handling of short initial frame that needs more data."""

    class MockMinFrameBleakClient(MockGobelBleakClient):
        """Send very short initial data then complete frame."""

        async def write_gatt_char(
            self,
            char_specifier: BleakGATTCharacteristic | int | str | UUID,
            data: Buffer,
            response: bool | None = None,
        ) -> None:
            """Issue write command and send short initial frame then rest."""
            await MockBleakClient.write_gatt_char(self, char_specifier, data, response)
            assert self._notify_callback is not None
            # First send a very short frame (less than MIN_FRAME_LEN=5)
            full_response = bytearray(self._RESP)
            self._notify_callback("MockMinFrameBleakClient", full_response[:3])
            # Then send the rest
            self._notify_callback("MockMinFrameBleakClient", full_response[3:])

    monkeypatch.setattr(MockMinFrameBleakClient, "_RESP", _FRAME_MAIN_DATA)
    patch_bleak_client(MockMinFrameBleakClient)

    bms = BMS(generate_ble_device())
    assert await bms.async_update() == _RESULT_MAIN_DATA

    await bms.disconnect()



"""Test helpers for Bleak / Bluetooth.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from typing import Any

from bleak.backends.device import BLEDevice


def generate_ble_device(
    address: str = "11:22:33:44:55:66",
    name: str | None = "MockBLEDevice",
    details: Any | None = None,
) -> BLEDevice:
    """Generate a BLEDevice with defaults."""
    return BLEDevice(address, name, details or {"path": ""})



"""Test the D-powercore BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.dpwrcore_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests


@pytest.fixture(
    name="dev_name",
    params=["TBA-MockBLEDevice_C0FE", "DXB-MockBLEDevice_C0FE", "invalid"],
    ids=["TBA", "DXB", "wrong"],
)
def patch_dev_name(request: pytest.FixtureRequest) -> str:
    """Provide device name variants."""
    assert isinstance(request.param, str)
    return request.param


_RESULT_DEFS: Final[BMSSample] = {
    "voltage": 53.1,
    "current": 0,
    "battery_level": 61.0,
    "cycles": 18,
    "cycle_charge": 17.806,
    "cell_voltages": [
        3.799,
        3.798,
        3.798,
        3.797,
        3.797,
        3.798,
        3.793,
        3.794,
        3.797,
        3.798,
        3.796,
        3.800,
        3.799,
        3.803,
    ],
    "cell_count": 14,
    "delta_voltage": 0.01,
    "temperature": 21.05,
    "temp_values": [21.05],
    "cycle_capacity": 945.499,
    "power": 0.0,
    "battery_charging": False,
    "problem": False,
    "problem_code": 0,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockDPwrcoreBleakClient(MockBleakClient):
    """Emulate a D-powercore BMS BleakClient."""

    PAGE_LEN = 20

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("fff3"):
            return bytearray()
        cmd: int = int(bytearray(data)[5])
        if cmd == 0x60:
            return bytearray(
                b"\x12\x12\x3a\x05\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\x00\x02\xcd"
                b"\x02\x22\x0d\x0a\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\x00\x02\xcd"
            )  # 2nd line only 4 bytes valid! TODO: put numbers
        if cmd == 0x61:
            return bytearray(
                b"\x12\x12\x3a\x05\x03\x61\x00\x0c\x00\x12\x00\x12\x6d\x60\x0b\x7e\x8f\xdb\x18\x20"
                b"\x04\x22\x03\x91\x0d\x0a\x00\x0c\x00\x12\x00\x12\x6d\x60\x0b\x7e\x8f\xdb\x18\x20"
            )  # 2nd line only 6 bytes valid! TODO: put numbers
        if cmd == 0x62:
            return bytearray(
                b"\x12\x13\x3a\x05\x03\x62\x00\x1d\x0e\x0e\xd7\x0e\xd6\x0e\xd6\x0e\xd5\x0e\xd5\x0e"
                b"\x12\x23\xd6\x0e\xd1\x0e\xd2\x0e\xd5\x0e\xd6\x0e\xd4\x0e\xd8\x0e\xd7\x0e\xdb\x0d"
                b"\x03\x33\x08\x0d\x0a\x0e\xd2\x0e\xd5\x0e\xd6\x0e\xd4\x0e\xd8\x0e\xd7\x0e\xdb\x0d"
            )  # 2nd line only 5 bytes valid TODO: put numbers
        if cmd == 0x64:
            assert bytearray(data)[8:10] == bytes.fromhex("C0FE"), "incorrect password"
            assert bytearray(data)[0] == 0xE, "incorrect unlock CMD length"
            resp = bytearray(data)
            return bytearray(resp[0] | 0x80) + resp[1:]
        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        data_ba = bytearray(data)
        await super().write_gatt_char(char_specifier, data, response)
        if data_ba[0] & 0x80:  # ignore ACK messages # TODO: verify those?
            return
        assert self._notify_callback is not None
        resp: bytearray = self._response(char_specifier, data)
        await self._notify_callback(  # send acknowledge
            "MockPwrcoreBleakClient", bytearray([data_ba[0] | 0x80]) + data_ba[1:]
        )
        for pos in range(1 + int((len(resp) - 1) / self.PAGE_LEN)):
            await self._notify_callback(
                "MockPwrcoreBleakClient", resp[pos * 20 :][: self.PAGE_LEN]
            )


class MockWrongCRCBleakClient(MockDPwrcoreBleakClient):
    """Emulate a D-powercore BMS BleakClient that replies with wrong CRC."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("fff3"):
            return bytearray()
        cmd: int = int(bytearray(data)[5])
        if cmd == 0x60:
            return bytearray(
                b"\x12\x12\x3a\x05\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\x00\x01\xce"
                b"\x02\x22\x0d\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            )  # wrong CRC [0x01CE != 0x02CD] in line 1
        if cmd == 0x61:
            return bytearray(
                b"\x12\x12\x3a\x05\x03\x61\x00\x0c\x00\x12\x00\x12\x6d\x60\x0b\x7e\x8f\xdb\x18\x20"
                b"\x04\x22\x02\x91\x0d\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            )  # wrong CRC [0x02 != 0x03] in line 2
        if cmd == 0x62:
            return bytearray(
                b"\x12\x13\x3a\x05\x03\x62\x00\x1d\x0e\x0e\xd7\x0e\xd6\x0e\xd6\x0e\xd5\x0e\xd5\x0e"
                b"\x12\x23\xd6\x0e\xd1\x0e\xd2\x0e\xd5\x0e\xd6\x0e\xd4\x0e\xd8\x0e\xd7\x0e\xdb\x0e"
                b"\x03\x33\x08\x0d\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            )  # wrong CRC [0x0E != 0x0D] in line 2
        return bytearray()


class MockInvalidBleakClient(MockDPwrcoreBleakClient):
    """Emulate a D-powercore BMS BleakClient replying garbage."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff3"):
            return bytearray(b"invalid_value")

        return bytearray()

    async def disconnect(self) -> None:
        """Mock disconnect to raise BleakError."""
        raise BleakError


class MockProblemBleakClient(MockDPwrcoreBleakClient):
    """Emulate a D-powercore BMS BleakClient reporting a problem."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("fff3"):
            return bytearray()
        if bytearray(data)[5] == 0x60:
            return bytearray(
                b"\x12\x12\x3a\x05\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\xff\x03\xcc"
                b"\x02\x22\x0d\x0a\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\x00\x03\xcc"
            )  # 2nd line only 4 bytes valid!

        return super()._response(char_specifier, data)


async def test_update(
    patch_bleak_client, dev_name: str, keep_alive_fixture: bool
) -> None:
    """Test D-pwercore BMS data update."""

    patch_bleak_client(MockDPwrcoreBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", dev_name), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_invalid_response(
    patch_bleak_client, patch_bms_timeout, dev_name: str
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()
    patch_bleak_client(MockInvalidBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", dev_name))

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


async def test_wrong_crc(patch_bleak_client, patch_bms_timeout, dev_name: str) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()
    patch_bleak_client(MockWrongCRCBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", dev_name))

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


async def test_problem_response(patch_bleak_client, dev_name: str) -> None:
    """Test D-pwercore BMS data update."""

    patch_bleak_client(MockProblemBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", dev_name), False)

    assert await bms.async_update() == _RESULT_DEFS | {
        "problem": True,
        "problem_code": 255,
    }

    await bms.disconnect()


async def test_incomplete_msgs(monkeypatch, patch_bleak_client, dev_name: str) -> None:
    """Test D-pwercore BMS data update."""

    def _stuck_response(
        _self,
        _char_specifier: BleakGATTCharacteristic | int | str | UUID,
        _data: Buffer,
    ) -> bytearray:
        return bytearray(
            b"\x12\x12\x3a\x05\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\x00\x02\xcd"
            b"\x02\x22\x0d\x0a\x03\x60\x00\x0a\x02\x13\x00\x00\x71\xc5\x45\x8e\x3d\x00\x02\xcd"
        )  # 2nd line only 4 bytes valid! TODO: put numbers

    monkeypatch.setattr(MockDPwrcoreBleakClient, "_response", _stuck_response)
    patch_bleak_client(MockDPwrcoreBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", dev_name), False)

    result: BMSSample = {}
    with pytest.raises(ValueError, match="BMS data incomplete."):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()



"""Module to support Gobel Power BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/

This module implements support for Gobel Power BMS devices that use Modbus RTU
protocol over Bluetooth Low Energy.
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, b2str, crc_modbus


class BMS(BaseBMS):
    """Gobel Power BLE BMS class implementation using Modbus RTU over BLE."""

    INFO: BMSInfo = {"default_manufacturer": "Gobel Power", "default_model": "BLE BMS"}

    # Modbus constants
    _SLAVE_ADDR: Final[int] = 0x01
    _FUNC_READ: Final[int] = 0x03

    # Frame constants
    _MIN_FRAME_LEN: Final[int] = 5  # addr + func + len + 2*crc minimum
    _MAX_CELLS: Final[int] = 32
    _MAX_TEMP: Final[int] = 8

    # Each command: (slave_addr, func_code, start_address, register_count)
    _RD_CMD_STATUS: Final[tuple[int, int, int, int]] = (
        _SLAVE_ADDR,
        _FUNC_READ,
        0x0000,
        0x003B,  # 59 registers
    )
    _RD_CMD_DEV_INFO: Final[tuple[int, int, int, int]] = (
        _SLAVE_ADDR,
        _FUNC_READ,
        0x00AA,
        0x0023,  # 35 registers
    )

    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("current", 0, 2, True, lambda x: x / 100),
        BMSDp("voltage", 2, 2, False, lambda x: x / 100),
        BMSDp("battery_level", 4, 2, False),
        BMSDp("battery_health", 6, 2, False),
        BMSDp("cycle_charge", 8, 2, False, lambda x: x / 100),
        BMSDp("design_capacity", 10, 2, False, lambda x: x // 100),
        BMSDp("cycles", 14, 2, False),
        BMSDp("problem_code", 16, 6, False),
        BMSDp("chrg_mosfet", 28, 2, False, lambda x: bool(x & 0x4000)),
        BMSDp("dischrg_mosfet", 28, 2, False, lambda x: bool(x & 0x8000)),
        BMSDp("cell_count", 30, 2, False, lambda x: x & 0xFF),
        BMSDp("temp_sensors", 92, 2, False, lambda x: x & 0xFF),
    )

    _CELLV_START: Final[int] = 35
    _TEMP_START: Final[int] = 97
    _TEMP_MOS_OFFSET: Final[int] = 117

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: bytes = b""

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [MatcherPattern(local_name="BMS-[0-9A-F]*", connectable=True)]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return ("00002760-08c2-11e1-9073-0e8ac72e1001",)

    @staticmethod
    def uuid_rx() -> str:
        """Return UUID of characteristic that provides notification/read property."""
        return "00002760-08c2-11e1-9073-0e8ac72e0002"

    @staticmethod
    def uuid_tx() -> str:
        """Return UUID of characteristic that provides write property."""
        return "00002760-08c2-11e1-9073-0e8ac72e0001"

    @staticmethod
    @cache
    def _cmd(addr: int, func: int, start: int, regs: int) -> bytes:
        """Build Modbus read command with CRC (cached)."""
        cmd: Final[bytes] = (
            addr.to_bytes(1, "big")
            + func.to_bytes(1, "big")
            + start.to_bytes(2, "big")
            + regs.to_bytes(2, "big")
        )
        return cmd + crc_modbus(cmd).to_bytes(2, "little")

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug(
            "RX BLE data (%s): %s", "start" if not self._frame else "cnt.", data
        )

        # Start of a new frame - check for valid Modbus response header
        if len(data) >= 2 and data[0] == BMS._SLAVE_ADDR:
            # Check if it's a valid read response or error response
            if data[1] == BMS._FUNC_READ or data[1] == (BMS._FUNC_READ | 0x80):
                # Start new frame (clear any old data)
                self._frame = bytearray(data)
            else:
                self._log.debug("unexpected function code: 0x%02X", data[1])
                return
        elif self._frame:
            # Continuation of existing frame
            self._frame.extend(data)
        else:
            self._log.debug("unexpected data, ignoring: %s", data.hex(" "))
            return

        # Check if we have enough data for minimum frame
        if len(self._frame) < BMS._MIN_FRAME_LEN:
            return

        # Check for error response
        if self._frame[1] == (BMS._FUNC_READ | 0x80):
            self._log.warning("Modbus error response: 0x%02X", self._frame[2])
            self._frame.clear()
            return

        # Get expected frame length from byte count field
        expected_len: Final[int] = BMS._MIN_FRAME_LEN + self._frame[2]

        if len(self._frame) < expected_len:
            return

        # Truncate if we received extra data
        del self._frame[expected_len:]

        # Verify CRC
        received_crc: Final[int] = int.from_bytes(self._frame[-2:], byteorder="little")
        calculated_crc: Final[int] = crc_modbus(self._frame[:-2])

        if received_crc != calculated_crc:
            self._log.debug(
                "invalid CRC: 0x%04X != 0x%04X", received_crc, calculated_crc
            )
            self._frame.clear()
            return

        self._log.debug("valid frame received: %d bytes", len(self._frame))
        self._msg = bytes(self._frame)
        self._msg_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_msg(BMS._cmd(*BMS._RD_CMD_STATUS))

        if self._msg[2] != BMS._RD_CMD_STATUS[3] * 2:
            self._log.debug(
                "incorrect response: %d bytes, expected %d",
                self._msg[2],
                BMS._RD_CMD_STATUS[3] * 2,
            )
            return {}

        result: BMSSample = BMS._decode_data(
            BMS._FIELDS, self._msg, byteorder="big", start=3
        )

        result["cell_voltages"] = BMS._cell_voltages(
            self._msg,
            cells=min(result.get("cell_count", 0), BMS._MAX_CELLS),
            start=BMS._CELLV_START,
            byteorder="big",
        )

        result["temp_values"] = BMS._temp_values(
            self._msg,
            values=min(result.get("temp_sensors", 0), BMS._MAX_TEMP),
            start=BMS._TEMP_START,
            byteorder="big",
            signed=True,
            divider=10,
        )

        # Append MOSFET temperature if valid (0xFFFF indicates no sensor)
        mos_temp: list[int | float] = BMS._temp_values(
            self._msg,
            values=1,
            start=BMS._TEMP_MOS_OFFSET,
            byteorder="big",
            signed=True,
            divider=10,
        )
        if mos_temp[0] != -0.1:
            result["temp_values"].append(mos_temp[0])

        return result

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch device info from BMS via Modbus."""
        # First get standard BLE device info (may contain generic values)
        info: BMSInfo = await super()._fetch_device_info()

        # Read device info registers via Modbus
        try:
            await self._await_msg(BMS._cmd(*BMS._RD_CMD_DEV_INFO))
        except TimeoutError:
            return info

        if len(self._msg) >= 65:
            info.update(
                {
                    "sw_version": b2str(self._msg[3:21]),
                    "serial_number": b2str(self._msg[23:43]),
                    "model_id": b2str(self._msg[43:63]),
                }
            )

        return info



"""Test the Vatrer BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.vatrer_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info


def ref_value() -> BMSSample:
    """Return reference value for mock Vatrer BMS."""
    return {
        "voltage": 52.67,
        "current": -4.96,
        "battery_level": 40,
        "battery_health": 100,
        "cycle_charge": 39.7,
        "cycle_capacity": 2090.999,
        "cycles": 27,
        "delta_voltage": 0.003,
        "cell_count": 16,
        "runtime": 28814,
        "temp_sensors": 4,
        "temp_values": [18.0, 20.0, 19.0, 20.0, 20.0, 20.0],
        "temperature": 19.5,
        "battery_charging": False,
        "power": -261.243,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "problem": False,
        "cell_voltages": [
            3.295,
            3.295,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.296,
            3.293,
        ],
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockVatrerBleakClient(MockBleakClient):
    """Emulate a Vatrer BMS BleakClient."""

    RESP: dict[bytes, bytearray] = {
        b"\x02\x03\x00\x34\x00\x12\x84\x3a": bytearray(
            b"\x02\x03\x24\x00\x04\x00\x12\x00\x14\x00\x13\x00\x14\x00\x14\x00\x14\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x34\x02\x40\x00\x00\x00\x00\x46\x43"
        ),  # temp info
        b"\x02\x03\x00\x00\x00\x14\x45\xf6": bytearray(
            b"\x02\x03\x28\x14\x93\xff\xff\xfe\x10\x00\x28\x0f\x82\x27\x10\x00\x1b\x00\x64\x00\x01"
            b"\x00\x00\x00\x01\x0c\xe0\x0c\xdd\x00\x03\x00\x0f\x00\x10\x00\x14\x00\x12\x00\x02\x00"
            b"\x04\xe4\xe5"
        ),  # status info
        b"\x02\x03\x00\x15\x00\x1f\x15\xf5": bytearray(
            b"\x02\x03\x3e\x00\x10\x0c\xdf\x0c\xdf\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0"
            b"\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xe0\x0c\xdd\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\xcb\x54"
        ),  # cell info
    }

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        self._notify_callback(
            "MockVatrerBleakClient", self.RESP.get(bytes(data), bytearray())
        )


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Vatrer BMS data update."""

    patch_bleak_client(MockVatrerBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockVatrerBleakClient, BMS)


@pytest.fixture(
    name="wrong_response",
    params=[
        (bytearray(b"\x01\x03\x24" + bytes(36) + b"\x7b\xa1"), "wrong_SOF"),
        (bytearray(b"\x02\x03\x24" + bytes(36) + b"\x60\x15\x00"), "wrong_length"),
        (bytearray(b"\x02\x03\x24" + bytes(36) + b"\x60\x16"), "wrong_CRC"),
        (bytearray(b"\x02\x03\x21" + bytes(33) + b"\xba\x66"), "wrong_type"),
        (bytearray(), "empty_frame"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockVatrerBleakClient,
        "RESP",
        MockVatrerBleakClient.RESP
        | {b"\x02\x03\x00\x34\x00\x12\x84\x3a": wrong_response},
    )

    patch_bleak_client(MockVatrerBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\x02\x03\x24\x00\x04\x00\x12\x00\x14\x00\x13\x00\x14\x00\x14\x00\x14\x80\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x34\x02\x40\x00\x00\x00\x00\xce\x2b"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\x02\x03\x24\x00\x04\x00\x12\x00\x14\x00\x13\x00\x14\x00\x14\x00\x14\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x34\x02\x40\x00\x00\x00\x00\x87\x8f"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockVatrerBleakClient,
        "RESP",
        MockVatrerBleakClient.RESP
        | {b"\x02\x03\x00\x34\x00\x12\x84\x3a": problem_response[0]},
    )

    patch_bleak_client(MockVatrerBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == ref_value() | {"problem": True}

    await bms.disconnect()



"""Example script for aiobmsble package usage.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import argparse
import asyncio
import getpass
import logging
from typing import Final

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.exc import BleakError

from aiobmsble import BMSInfo, BMSSample, __version__
from aiobmsble.basebms import BaseBMS
from aiobmsble.utils import bms_identify

logging.basicConfig(
    format="%(levelname)s: %(message)s",
    level=logging.INFO,
)
logger: logging.Logger = logging.getLogger(__package__)


async def scan_devices() -> dict[str, tuple[BLEDevice, AdvertisementData]]:
    """Scan for BLE devices and return results."""
    logger.info("starting scan ...")
    try:
        scan_result: dict[str, tuple[BLEDevice, AdvertisementData]] = (
            await BleakScanner.discover(return_adv=True)
        )
    except BleakError as exc:
        logger.error("Could not scan for BT devices: %s", exc)
        return {}

    logger.debug(scan_result)
    logger.info("%i BT device(s) in range.", len(scan_result))
    return scan_result


async def _try_query(
    bms_cls: type[BaseBMS], ble_dev: BLEDevice, secret: str = ""
) -> bool:
    """Attempt to query the BMS once. Returns True on success, False on failure."""
    bms_inst: BaseBMS = (
        bms_cls(ble_device=ble_dev, secret=secret)
        if secret
        else bms_cls(ble_device=ble_dev)
    )
    logger.info("Querying BMS%s...", " with secret" if secret else "")

    try:
        async with bms_inst as bms:
            info: BMSInfo = await bms.device_info()
            data: BMSSample = await bms.async_update()
        logger.info("BMS info: %s", repr(info).replace(", '", ",\n\t'"))
        logger.info("BMS data: %s", repr(data).replace(", '", ",\n\t'"))
    except (BleakError, TimeoutError) as exc:
        logger.error(
            "Failed to query BMS%s: %s",
            " with secret" if secret else "",
            type(exc).__name__,
        )
        return False
    return True


async def detect_bms() -> None:
    """Query a Bluetooth device based on the provided arguments."""

    scan_result: dict[str, tuple[BLEDevice, AdvertisementData]] = await scan_devices()
    for ble_dev, advertisement in scan_result.values():
        logger.info(
            "%s\nBT device '%s' (%s)\n\t%s",
            "-" * 72,
            ble_dev.name,
            ble_dev.address,
            repr(advertisement).replace(", ", ",\n\t"),
        )

        if bms_cls := await bms_identify(advertisement, ble_dev.address):
            bms_inst: BaseBMS = bms_cls(ble_device=ble_dev)
            logger.info("Found matching BMS type: %s", bms_inst.bms_id())

            if not await _try_query(bms_cls, ble_dev):
                if bms_cls.accept_secret:
                    secret: str = getpass.getpass(
                        f"Enter secret for {bms_cls.__name__}: "
                    )
                    await _try_query(bms_cls, ble_dev, secret)

    logger.info("done.")


def setup_logging(args: argparse.Namespace) -> None:
    """Configure logging based on command line arguments."""
    loglevel: Final[int] = logging.DEBUG if args.verbose else logging.INFO

    if args.logfile:
        file_handler = logging.FileHandler(args.logfile)
        file_handler.setLevel(loglevel)
        file_handler.setFormatter(
            logging.Formatter("%(asctime)s - %(levelname)s: %(message)s")
        )
        logger.addHandler(file_handler)

    logger.setLevel(loglevel)
    logger.info("%s version %s", __package__, __version__)


def main() -> None:
    """Entry point for the script to run the BMS detection."""
    parser = argparse.ArgumentParser(
        description="Reference script for 'aiobmsble' to show all recognized BMS in range."
    )
    parser.add_argument("-l", "--logfile", type=str, help="Path to the log file")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable debug logging"
    )

    setup_logging(parser.parse_args())

    asyncio.run(detect_bms())


if __name__ == "__main__":
    main()  # pragma: no cover



"""Test the EG4 BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble.basebms import BMSSample
from aiobmsble.bms.eg4_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE: Final[int] = 80
CMD_STATUS: Final[bytes] = b"\x01\x03\x00\x00\x00\x27\x05\xd0"
CMD_INFO: Final[bytes] = b"\x01\x03\x00\x69\x00\x2e\x15\xca"

_RESULT_DEFS: Final[BMSSample] = {
    "balancer": 0,
    "voltage": 13.24,
    "current": -6.2,
    "cycles": 1,  # not in recording
    "battery_health": 100,
    "battery_level": 48,
    "cell_count": 4,
    "cell_voltages": [3.31, 3.313, 3.309, 3.313],
    "delta_voltage": 0.004,
    "temperature": 17,
    "power": -82.088,
    "battery_charging": False,
    "problem_code": 0,
    "problem": False,
    "temp_values": [17.0, -56.0, 0.0, 0.0, 0.0, 0.0],
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockEG4BleakClient(MockBleakClient):
    """Emulate a EG4 BMS BleakClient."""

    _TX_CHAR_UUID: Final[str] = "1001"
    _RESP: dict[bytes, bytearray] = {
        CMD_STATUS: bytearray(
            b"\x01\x03\x4e\x05\x2c\xff\xc2\x0c\xee\x0c\xf1\x0c\xed\x0c\xf1\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x00"
            b"\x11\x00\x11\x00\x00\x02\xee\x00\x64\x00\x30\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x01\x00\x00\x00\x00\x11\xc8\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x95\xa4"  # \xc4\x34"
        ),  # 13.24V, -6.2A, 17°C, 48%, dischrg, ver 1.5(7), 3.309V, 3.314V, 3.309V, 3.312V
        CMD_INFO: bytearray(
            b"\x01\x03\x2e\x32\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x30"
            b"\x36\x00\x00\x00\x00\x30"
            b"\x31\x36\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x97\x8c"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        """Generate response based on command."""
        if not isinstance(char_specifier, str) or normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str(self._TX_CHAR_UUID):
            return bytearray()

        return self._RESP.get(bytes(data), bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        for notify_data in [
            self._response(char_specifier, data)[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._response(char_specifier, data)), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockEG4BleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test EG4 BMS data update."""

    patch_bleak_client(MockEG4BleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms._client and bms._client.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockEG4BleakClient, BMS)


@pytest.mark.parametrize(
    "wrong_response",
    [
        b"\x02" + MockEG4BleakClient._RESP[CMD_STATUS][1:],  # wrong_SOF
        MockEG4BleakClient._RESP[CMD_STATUS][:-2] + b"\xe1\x65",  # wrong_CRC
        b"\x01",
        b"",
    ],
    ids=["wrong_SOF", "wrong_CRC", "too_short", "empty"],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockEG4BleakClient,
        "_RESP",
        MockEG4BleakClient._RESP | {CMD_STATUS: bytearray(wrong_response)},
    )
    patch_bleak_client(MockEG4BleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.mark.parametrize(
    "problem_response",
    [
        b"\x01\x03\x4e\x05\x2c\xff\xc2\x0c\xee\x0c\xf1\x0c\xed\x0c\xf1\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x00\x11\x00"
        b"\x11\x00\x00\x02\xee\x00\x64\x00\x30\x00\x02\x80\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00"
        b"\x00\x00\x00\x11\xc8\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x3f\xd8",  # \x6e\x48
        b"\x01\x03\x4e\x05\x2c\xff\xc2\x0c\xee\x0c\xf1\x0c\xed\x0c\xf1\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x00\x11\x00"
        b"\x11\x00\x00\x02\xee\x00\x64\x00\x30\x00\x02\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00"
        b"\x00\x00\x00\x11\xc8\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x85\x75",  # \xd4\xe5",
    ],
    ids=["last_bit", "first_bit"],
)
async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: bytes,
    request,
) -> None:
    """Test data update with BMS returning error flags."""
    monkeypatch.setattr(
        MockEG4BleakClient,
        "_RESP",
        MockEG4BleakClient._RESP
        | {b"\x01\x03\x00\x00\x00\x27\x05\xd0": bytearray(problem_response)},
    )
    patch_bleak_client(MockEG4BleakClient)
    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS | {
        "problem": True,
        "problem_code": (1 if request.node.callspec.id == "first_bit" else 2**47),
    }

    await bms.disconnect()



"""Module to support JBD smart BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, b2str, crc_sum, swap32


class BMS(BaseBMS):
    """JBD smart BMS class implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Jiabaida", "default_model": "smart BMS"}
    _HEAD_INIT: Final[bytes] = b"\xff\xaa"  # header for initialization
    _HEAD_RSP: Final[bytes] = b"\xdd"  # header for responses
    _HEAD_CMD: Final[bytes] = b"\xdd\xa5"  # read header for commands
    _TAIL: Final[int] = 0x77  # tail for command
    _INIT_LEN: Final[int] = 5  # initialization frame size
    _INFO_LEN: Final[int] = 7  # minimum frame size
    _BASIC_INFO: Final[int] = 23  # basic info data length
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 4, 2, False, lambda x: x / 100),
        BMSDp("current", 6, 2, True, lambda x: x / 100),
        BMSDp("cycle_charge", 8, 2, False, lambda x: x / 100),
        BMSDp("design_capacity", 10, 2, False, lambda x: x // 100),
        BMSDp("cycles", 12, 2, False),
        BMSDp("balancer", 16, 4, False, swap32),
        BMSDp("problem_code", 20, 2, False),
        BMSDp("battery_level", 23, 1, False),
        BMSDp("chrg_mosfet", 24, 1, False, lambda x: bool(x & 0x1)),
        BMSDp("dischrg_mosfet", 24, 1, False, lambda x: bool(x & 0x2)),
        BMSDp("temp_sensors", 26, 1, False),  # count is not limited
    )  # general protocol v4

    accept_secret: bool = True

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._secret: Final[str] = secret
        self._valid_reply: int = 0x00
        self._msg: bytes = b""

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            MatcherPattern(
                local_name=pattern,
                service_uuid=BMS.uuid_services()[0],
                connectable=True,
            )
            for pattern in (
                "JBD-*",
                "SX1*",  # Supervolt v3
                "SX60*",  # Supervolt Ultra
                "SBL-*",  # SBL
                "OGR-*",  # OGRPHY
                "TZ-H*",  # CERRNSS battery
            )
        ] + [
            MatcherPattern(
                oui=oui, service_uuid=BMS.uuid_services()[0], connectable=True
            )
            for oui in (
                "10:A5:62",  # CHINS
                "A4:C1:37",
                "A4:C1:38",
                "A5:C2:37",
                "A5:C2:39",
                "AA:C2:37",
                "70:3E:97",
            )
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ff00"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ff01"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ff02"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        await self._await_cmd_resp(0x05)
        length: Final[int] = self._msg[3]
        return {
            "hw_version": b2str(self._msg[4 : length + 4]),
        }

    def _notify_init_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new init data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(BMS._HEAD_INIT):
            self._log.debug("incorrect SOF")
            return

        if len(data) < BMS._INIT_LEN or len(data) < BMS._INIT_LEN + data[3]:
            self._log.debug("incorrect frame length")
            return

        if (crc := crc_sum(data[2:-1])) != data[-1]:
            self._log.debug("invalid checksum 0x%X != 0x%X", data[-1], crc)
            return

        if data[2] != 0x15 or data[3] != 0x01:
            self._log.debug("unexpected response")
            return

        self._msg = bytes(data)
        self._msg_event.set()

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        if self._secret:
            await self._client.start_notify(BMS.uuid_rx(), self._notify_init_handler)
            data: Final[bytes] = self._secret.encode(encoding="ASCII")
            try:
                await self._await_msg(
                    BMS._HEAD_INIT
                    + b"\x15"
                    + len(self._secret).to_bytes(1)
                    + data
                    + ((0x15 + len(self._secret) + sum(data)) & 0xFF).to_bytes(1)
                )
            except TimeoutError:
                self._log.warning("Failed to initialize connection with secret.")
                raise
            if self._msg[4] != 0x00:
                self._log.warning("Incorrect secret.")
                raise PermissionError("Incorrect secret.")

            await self._client.stop_notify(BMS.uuid_rx())

        await super()._init_connection(char_notify)

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        # check if answer is a heading of basic info (0x3) or cell block info (0x4)
        if (
            data.startswith(BMS._HEAD_RSP)
            and len(self._frame) > BMS._INFO_LEN
            and data[1] in (0x03, 0x04, 0x05)
            and data[2] == 0x00
            and len(self._frame) >= BMS._INFO_LEN + self._frame[3]
        ):
            self._frame.clear()

        self._frame += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        # verify that data is long enough
        if (
            len(self._frame) < BMS._INFO_LEN
            or len(self._frame) < BMS._INFO_LEN + self._frame[3]
        ):
            return

        # check correct frame ending
        frame_end: Final[int] = BMS._INFO_LEN + self._frame[3] - 1
        if self._frame[frame_end] != BMS._TAIL:
            self._log.debug("incorrect frame end (length: %i).", len(self._frame))
            return

        if (crc := BMS._crc(self._frame[2 : frame_end - 2])) != int.from_bytes(
            self._frame[frame_end - 2 : frame_end], "big"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(self._frame[frame_end - 2 : frame_end], "big"),
                crc,
            )
            return

        if len(self._frame) != BMS._INFO_LEN + self._frame[3]:
            self._log.debug("wrong data length (%i): %s", len(self._frame), self._frame)

        if self._frame[1] != self._valid_reply:
            self._log.debug("unexpected response (type 0x%X)", self._frame[1])
            return

        self._msg = bytes(self._frame)
        self._msg_event.set()

    @staticmethod
    def _crc(frame: bytearray) -> int:
        """Calculate JBD frame CRC."""
        return 0x10000 - sum(frame)

    @staticmethod
    @cache
    def _cmd(cmd: int, data: bytes = b"") -> bytes:
        """Assemble a JBD BMS command."""
        frame = bytearray(
            BMS._HEAD_CMD + cmd.to_bytes(1) + len(data).to_bytes(1) + data
        )
        frame.extend([*BMS._crc(frame[2:]).to_bytes(2, "big"), BMS._TAIL])
        return bytes(frame)

    async def _await_cmd_resp(self, cmd: int) -> None:
        msg: Final[bytes] = BMS._cmd(cmd)
        self._valid_reply = msg[2]
        await self._await_msg(msg)
        self._valid_reply = 0x00

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        data: BMSSample = {}
        await self._await_cmd_resp(0x03)
        data = BMS._decode_data(BMS._FIELDS, self._msg)
        data["temp_values"] = BMS._temp_values(
            self._msg,
            values=data.get("temp_sensors", 0),
            start=27,
            signed=False,
            offset=2731,
            divider=10,
        )

        await self._await_cmd_resp(0x04)
        data["cell_voltages"] = BMS._cell_voltages(
            self._msg, cells=self._msg[3] // 2, start=4, byteorder="big"
        )

        return data



"""Test the Daly BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.basebms import crc_modbus
from aiobmsble.bms.daly_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests


def ref_value() -> BMSSample:
    """Return reference value for mock Daly BMS."""
    return {
        "voltage": 14.0,
        "current": 3.0,
        "battery_level": 90.0,
        "cycles": 57,
        "cycle_charge": 345.6,
        "cell_voltages": [4.127, 4.137, 4.147, 4.157],
        "cell_count": 4,
        "delta_voltage": 0.321,
        "temp_sensors": 4,
        "cycle_capacity": 4838.4,
        "power": 42.0,
        "battery_charging": True,
        "problem": False,
        "problem_code": 0,
        "chrg_mosfet": False,
        "dischrg_mosfet": True,
        "balancer": True,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockDalyBleakClient(MockBleakClient):
    """Emulate a Daly BMS BleakClient."""

    HEAD_READ: Final[bytes] = b"\xd2\x03"
    CMD_INFO: Final[bytes] = b"\x00\x00\x00\x3e\xd7\xb9"
    MOS_INFO: Final[bytes] = b"\x00\x3e\x00\x09\xf7\xa3"
    VER_INFO: Final[bytes] = b"\x00\xa9\x00\x20\x87\x91"
    MOS_AVAIL: bool = True
    RESP: Final[dict[bytes, bytearray]] = {
        CMD_INFO: bytearray(
            b"\xd2\x03\x7c\x10\x1f\x10\x29\x10\x33\x10\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x3d\x00\x3e\x00\x3f\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x8c\x75\x4e\x03\x84\x10\x3d\x10\x1f\x00\x00\x00\x00\x00\x00\x0d"
            b"\x80\x00\x04\x00\x04\x00\x39\x00\x01\x00\x00\x00\x01\x10\x2e\x01\x41\x00\x2a\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\xa0\xdf"
        ),  # 'voltage': 14.0, 'current': 3.0, 'battery_level': 90.0, 'cycles': 57,
        # 'cycle_charge': 345.6, 'numTemp': 4, 'temperature': 21.5, 'cycle_capacity': 4838.400000000001,
        # 'power': 42.0, 'battery_charging': True, 'runtime': none!, 'delta_voltage': 0.321
        MOS_INFO: bytearray(
            b"\xd2\x03\x12\x00\x00\x00\x00\x75\x30\x00\x00\x00\x4e\xff\xff\xff\xff\xff\xff\xff"
            b"\xff\x0b\x4e"
        ),
        VER_INFO: bytearray(
            b"\xd2\x03\x40\x54\x30\x30\x4b\x5f\x33\x32\x31\x30\x34\x32\x5f\x31\x31\x00\x00\x48"
            b"\x32\x2e\x30\x5f\x31\x30\x33\x52\x5f\x33\x30\x39\x46\x39\x46\x32\x30\x32\x34\x30"
            b"\x32\x32\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x55\x41"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("fff2")
            and bytes(data)[0:2] == self.HEAD_READ
        ):
            if bytes(data)[2:] == self.MOS_INFO and not self.MOS_AVAIL:
                raise TimeoutError
            return MockDalyBleakClient.RESP.get(bytes(data)[2:], bytearray())

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        self._notify_callback(
            "MockDalyBleakClient", self._response(char_specifier, data)
        )


class MockInvalidBleakClient(MockDalyBleakClient):
    """Emulate a Daly BMS BleakClient."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff2"):
            return bytearray(
                b"\xd2\x03\x11\x10\x1f\x10\x29\x10\x33\x10\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x5d\x0f"
            )

        return bytearray()

    async def disconnect(self) -> None:
        """Mock disconnect to raise BleakError."""
        raise BleakError


@pytest.mark.parametrize("mos_sensor_avail", [True, False])
async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    mos_sensor_avail: bool,
    keep_alive_fixture: bool,
) -> None:
    """Test Daly BMS data update."""

    monkeypatch.setattr(  # patch recoginiation of MOS request to fail
        MockDalyBleakClient, "MOS_AVAIL", mos_sensor_avail
    )

    patch_bleak_client(MockDalyBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value() | (
        {
            "temperature": 24.8,
            "temp_values": [38.0, 20.0, 21.0, 22.0, 23.0],
        }
        if mos_sensor_avail
        else {
            "temperature": 21.5,
            "temp_values": [20.0, 21.0, 22.0, 23.0],
        }
    )

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockDalyBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "hw_version": "H2.0_103R_309F9F",
        "sw_version": "T00K_321042_11",
    }


@pytest.mark.parametrize(
    "test_seq",
    [
        # test ignore invalid MOS after valid read
        ((b"\x00\x30", [8.0]), (b"\x00\x00", [-40.0])),
        ((b"\x00\x30", [8.0]), (b"\xff\xff", [65495.0])),
        # test disabling of MOS read after initial invalid value
        ((b"\xff\xff", []), (b"\x00\x30", [])),
        ((b"\x00\x00", []), (b"\x00\x30", [])),
    ],
)
async def test_mos_excl(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    test_seq: tuple[tuple[bytearray, list[int | float]], ...],
) -> None:
    """Test Daly BMS data update."""

    patch_bleak_client(MockDalyBleakClient)
    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", "MockBLEdevice"))

    for response, expected in test_seq:
        mos_info: bytearray = MockDalyBleakClient.RESP[
            MockDalyBleakClient.MOS_INFO
        ].copy()
        mos_info[BMS._MOSTEMP_POS : BMS._MOSTEMP_POS + 2] = response
        mos_info[-2:] = crc_modbus(mos_info[:-2]).to_bytes(2, byteorder="little")
        monkeypatch.setattr(
            MockDalyBleakClient,
            "RESP",
            MockDalyBleakClient.RESP | {MockDalyBleakClient.MOS_INFO: mos_info},
        )
        assert await bms.async_update() == ref_value() | {
            "temperature": (sum(expected) + 86) / (len(expected) + 4),
            "temp_values": [*expected, 20.0, 21.0, 22.0, 23.0],
        }

    await bms.disconnect()


async def test_too_short_frame(patch_bleak_client) -> None:
    """Test data update with BMS returning valid but too short data."""

    patch_bleak_client(MockInvalidBleakClient)

    bms: BMS = BMS(generate_ble_device())

    assert not await bms.async_update()

    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (bytearray(b"invalid_value"), "invalid value"),
        (
            bytearray(
                b"\xd2\x03\x7c\x10\x1f\x10\x29\x10\x33\x10\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x3d\x00\x3e\x00\x3f\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x8c\x75\x4e\x03\x84\x10\x3d\x10\x1f\x00\x00\x00\x00\x00\x00\x0d"
                b"\x80\x00\x04\x00\x04\x00\x39\x00\x01\x00\x00\x00\x01\x10\x2e\x01\x41\x00\x2a\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\xde\xad"
            ),
            "wrong CRC",
        ),
        (bytearray(b"\x00"), "too short"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockDalyBleakClient, "_response", lambda _s, _c, _d: wrong_response
    )

    patch_bleak_client(MockDalyBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\xd2\x03\x7c\x10\x1f\x10\x29\x10\x33\x10\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x3d\x00\x3e\x00\x3f\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x8c\x75\x4e\x03\x84\x10\x3d\x10\x1f\x00\x00\x00\x00\x00\x00\x0d"
                b"\x80\x00\x04\x00\x04\x00\x39\x00\x01\x00\x00\x00\x01\x10\x2e\x01\x41\x00\x2a\x00"
                b"\x00\x00\x00\x00\x00\x00\x01\x61\x1f"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\xd2\x03\x7c\x10\x1f\x10\x29\x10\x33\x10\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x3d\x00\x3e\x00\x3f\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x8c\x75\x4e\x03\x84\x10\x3d\x10\x1f\x00\x00\x00\x00\x00\x00\x0d"
                b"\x80\x00\x04\x00\x04\x00\x39\x00\x01\x00\x00\x00\x01\x10\x2e\x01\x41\x00\x2a\x80"
                b"\x00\x00\x00\x00\x00\x00\x00\xa8\xbf"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest):
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockDalyBleakClient, "_response", lambda _s, _c, _d: problem_response[0]
    )

    patch_bleak_client(MockDalyBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result.get("problem", False)  # we expect a problem
    assert result.get("problem_code", 0) == (
        1 << (0 if problem_response[1] == "first_bit" else 63)
    )

    await bms.disconnect()



"""Test the Pace BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.pace_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 200


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockPaceBleakClient(MockBleakClient):
    """Emulate a Pace BMS BleakClient."""

    _RESP: Final[dict[bytes, bytearray]] = {
        b"\x9a\x00\x00\x00\x01\x00\x00\x00\xe4\xc8\x9d": bytearray(  # REQ not in traces!
            b"\x9a\x00\x00\x00\x01\x00\x00\x43\x00\x15\x50\x31\x36\x53\x31\x30"
            b"\x30\x41\x2d\x33\x31\x39\x31\x36\x2d\x31\x2e\x30\x36\x00\x00\x20"
            b"\x50\x31\x36\x53\x31\x30\x30\x41\x2d\x33\x31\x39\x31\x36\x2d\x31"
            b"\x2e\x30\x36\x2d\x30\x30\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x0a\x32\x31\x39\x33\x39\x56\x31\x34\x30\x00\x6c\xa7\x9d"
        ),
        b"\x9a\x00\x00\x00\x04\x00\x00\x00\x28\xc8\x9d": bytearray(
            b"\x9a\x00\x00\x00\x04\x00\x00\x17\x01\x15\x34\x33\x32\x31\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa1"
            b"\x9b\x9d"
        ),
        b"\x9a\x00\x00\x0a\x00\x00\x00\x00\x19\x51\x9d": bytearray(
            b"\x9a\x00\x00\x0a\x00\x00\x00\x33\x01\xff\xff\xfe\x59\x00\x00\x14"
            b"\xb5\x00\x00\x1c\xe8\x00\x00\x27\x10\x00\x00\x27\x10\x4a\x64\x00"
            b"\x00\x00\xef\x00\x00\x00\x00\x00\x00\x00\x00\x01\x08\x0c\xef\x01"
            b"\x10\x0c\xec\x01\x03\x0b\x8e\x01\x01\x0b\x89\x99\x16\x9d"
        ),  # system values: voltage 53.01, current -4.23, battery level 74, cycle: 239
        b"\x9a\x00\x00\x0a\x03\x00\x00\x02\x01\x01\xf9\x9d\x9d": bytearray(
            b"\x9a\x00\x00\x0a\x03\x00\x00\x0d\x02\x00\x00\x0e\x06\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x18\xae\x9d"
        ),
        b"\x9a\x00\x00\x00\x07\x00\x00\x00\x6c\xc8\x9d": bytearray(
            b"\x9a\x00\x00\x00\x07\x00\x00\x06\x19\x0a\x12\x11\x0c\x14\x3b\x2c\x9d"
        ),  # date/time: 25-10-18 17:12:20
        b"\x9a\x00\x00\x00\x05\x00\x00\x00\xd4\xc9\x9d": bytearray(
            b"\x9a\x00\x00\x00\x05\x00\x00\x16\x15\x34\x33\x32\x31\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\xe8\x9d"
        ),
        b"\x9a\x00\x00\x00\x0c\x00\x00\x00\x48\xca\x9d": bytearray(
            b"\x9a\x00\x00\x00\x0c\x00\x00\x3e\x12\x00\x01\x00\x02\x00\x03\x00"
            b"\x04\x00\x05\x00\x06\x00\x07\x00\x08\x00\x09\x00\x0a\x00\x0b\x00"
            b"\x0c\x00\x0e\x00\x0f\x00\x24\x00\x13\x00\x3d\x00\x3e\x0c\x00\x01"
            b"\x00\x02\x00\x03\x00\x04\x00\x0c\x00\x0d\x00\x28\x00\x0f\x00\x13"
            b"\x00\x2b\x00\x3d\x00\x3e\xf4\x08\x9d"
        ),
        b"\x9a\x00\x00\x00\x0d\x00\x00\x00\xb4\xcb\x9d": bytearray(
            b"\x9a\x00\x00\x00\x0d\x00\x00\x04\x00\x0b\x00\x3e\xd0\x7f\x9d"
        ),
        b"\x9a\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x4b\x6c\x0e\x9d": bytearray(
            b"\x9a\x00\x00\x00\x00\x00\x00\x9a\x00\x00\x00\x4b\x00\x01\x1c\xff"
            b"\x27\x10\x27\x10\x00\x4a\xde\xa8\x07\xd0\x00\x0a\x00\x10\x0d\x16"
            b"\x00\x1e\x00\xef\x0b\xea\x00\x0a\x00\x0c\x00\x64\x00\x01\x00\x0a"
            b"\x16\x80\x16\xd0\x15\x18\x00\x01\x00\x0a\x12\xc0\x10\xe0\x12\x20"
            b"\x00\x01\x00\x0a\x0e\x10\x0e\x42\x0d\x34\x00\x01\x00\x0a\x0b\xb8"
            b"\x0a\x8c\x0b\x54\x00\x01\x00\x0a\x00\x69\x00\x6e\x00\x00\x00\x01"
            b"\x00\x0a\x00\x69\x00\x6e\x00\x00\x00\x96\x00\x05\x00\x96\x00\x05"
            b"\x00\x01\x0c\x9e\x0c\xd0\x0c\x9e\x0c\xd0\x0d\x02\x0c\xd0\x00\x01"
            b"\x0a\xc8\x0a\x96\x0a\xc8\x0a\x14\x09\xe2\x0a\x14\x00\x01\x0e\x2e"
            b"\x0e\xc4\x0d\xfc\x00\x01\x09\xe2\x09\xb0\x09\xe2\x0d\x34\x0d\x66"
            b"\x0d\x34\xfe\x74\x9d"
        ),
        b"\x9a\x00\x00\x00\x00\x00\x00\x04\xf1\x01\x00\x07\xa5\x6c\x9d": bytearray(
            b"\x9a\x00\x00\x00\x00\x00\x00\x12\xf1\x01\x00\x07\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x83\xac\x9d"
        ),
        b"\x9a\x00\x00\x00\x02\x00\x00\x00\xa0\xc8\x9d": bytearray(
            b"\x9a\x00\x00\x00\x02\x00\x00\x0e\x0d\x31\x32\x33\x34\x35\x36\x37"
            b"\x38\x39\x30\x31\x32\x5a\x7d\x7d\x9d"
        ),  # serial number 123456789012Z
        b"\x9a\x00\x00\x0a\x01\x00\x00\x02\x01\x01\x1b\x9c\x9d": bytearray(
            b"\x9a\x00\x00\x0a\x01\x00\x00\x2b\x02\xfe\x80\x14\xb6\x1c\xff\x27"
            b"\x10\x27\x10\x4a\x64\x00\xef\x00\x00\x00\x00\x00\x00\x00\x00\x14"
            b"\xd7\x00\x00\x03\x0b\x8e\x01\x0b\x89\x0b\x93\x0b\x9c\x0a\x0c\xf0"
            b"\x10\x0c\xec\xa4\xb4\x9d"
        ),  # pack -3.84A, 53.02V
        b"\x9a\x00\x00\x0a\x02\x00\x00\x02\x01\x01\x28\x9c\x9d": bytearray(
            b"\x9a\x00\x00\x0a\x02\x00\x00\x44\x02\x14\xb6\x10\x0c\xee\x0b\x89"
            b"\x0c\xec\x0b\x8b\x0c\xee\x0b\x8e\x0c\xef\x0b\x8b\x0c\xef\x00\x00"
            b"\x0c\xee\x00\x00\x0c\xef\x00\x00\x0c\xef\x00\x00\x0c\xef\x00\x00"
            b"\x0c\xf0\x00\x00\x0c\xef\x00\x00\x0c\xee\x00\x00\x0c\xee\x00\x00"
            b"\x0c\xef\x00\x00\x0c\xee\x00\x00\x0c\xec\x00\x00\x32\xa1\x9d"
        ),  # pack 53.02V
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        frame: Final[bytes] = bytes(data)
        assert char_specifier == "fff2"

        return self._RESP.get(frame, bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: Final[bytearray] = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockPaceBleakClient", notify_data)


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockPaceBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "serial_number": "123456789012Z",
        "sw_version": "P16S100A-31916-1.06",
        "hw_version": "21939V140",
    }


async def test_update(patch_bleak_client, keep_alive_fixture) -> None:
    """Test Dummy BMS data update."""

    patch_bleak_client(MockPaceBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == {
        "pack_count": 1,
        "voltage": 53.01,
        "current": -4.23,
        "battery_level": 74,
        "battery_health": 100,
        "cycle_charge": 74.0,
        "design_capacity": 100,
        "cycle_capacity": 3922.74,
        "delta_voltage": 0.004,
        "cycles": 239,
        "cell_count": 16,
        "cell_voltages": [
            3.31,
            3.308,
            3.31,
            3.311,
            3.311,
            3.31,
            3.311,
            3.311,
            3.311,
            3.312,
            3.311,
            3.31,
            3.31,
            3.311,
            3.31,
            3.308,
        ],
        "temp_values": [22.2, 22.4, 22.7, 22.4],
        "temperature": 22.425,
        "battery_charging": False,
        "runtime": 62978,
        "power": -224.232,
        "problem": False,
    }

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"\x90\x00\x00\x0a\x00\x00\x00\x33\x01\xff\xff\xfe\x59\x00\x00\x14\xb5\x00\x00\x1c\xe8\x00"
        b"\x00\x27\x10\x00\x00\x27\x10\x4a\x64\x00\x00\x00\xef\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        b"\x08\x0c\xef\x01\x10\x0c\xec\x01\x03\x0b\x8e\x01\x01\x0b\x89\x91\x36\x9d",
        b"\x9a\x00\x00\x0a\x00\x00\x00\x33\x01\xff\xff\xfe\x59\x00\x00\x14\xb5\x00\x00\x1c\xe8\x00"
        b"\x00\x27\x10\x00\x00\x27\x10\x4a\x64\x00\x00\x00\xef\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        b"\x08\x0c\xef\x01\x10\x0c\xec\x01\x03\x0b\x8e\x01\x01\x0b\x89\x99\x99\x9d",
        b"\x9a\x00\x00\x0a\x00\x00\x00\x33\x01\xff\xff\xfe\x59\x00\x00\x14\xb5\x00\x00\x1c\xe8\x00"
        b"\x00\x27\x10\x00\x00\x27\x10\x4a\x64\x00\x00\x00\xef\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        b"\x08\x0c\xef\x01\x10\x0c\xec\x01\x03\x0b\x8e\x01\x01\x0b\x89\x00\x16\x11\x9d",
        b"\x9a\x00\x00\x00\x00\x00\x00\x33\x01\xff\xff\xfe\x59\x00\x00\x14\xb5\x00\x00\x1c\xe8\x00"
        b"\x00\x27\x10\x00\x00\x27\x10\x4a\x64\x00\x00\x00\xef\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        b"\x08\x0c\xef\x01\x10\x0c\xec\x01\x03\x0b\x8e\x01\x01\x0b\x89\x59\x3d\x9d",
        b"",
    ],
    ids=["wrong_SOF", "wrong_CRC", "wrong_len", "wrong_response", "empty"],
)
async def test_invalid_response(
    monkeypatch, patch_bleak_client, patch_bms_timeout, wrong_response: bytes
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockPaceBleakClient,
        "_RESP",
        MockPaceBleakClient._RESP
        | {b"\x9a\x00\x00\x0a\x00\x00\x00\x00\x19\x51\x9d": bytearray(wrong_response)},
    )
    patch_bleak_client(MockPaceBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()



"""Module to support Neey smart BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Callable
from functools import cache
from struct import unpack_from
from typing import Any, Final, Literal

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, b2str, crc_sum


class BMS(BaseBMS):
    """Neey smart BMS class implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Neey", "default_model": "Balancer"}
    _BT_MODULE_MSG: Final = b"\x41\x54\x0d\x0a"  # AT\r\n from BLE module
    _HEAD_RSP: Final = b"\x55\xaa\x11\x01"  # start, dev addr, read cmd
    _HEAD_CMD: Final = b"\xaa\x55\x11\x01"  # cmd header (endianness!)
    _TAIL: Final[int] = 0xFF  # end of message
    _TYPE_POS: Final[int] = 4  # frame type is right after the header
    _MIN_FRAME: Final[int] = 10  # header length
    _FIELDS: Final[tuple[tuple[BMSValue, int, str, Callable[[int], Any]], ...]] = (
        ("voltage", 201, "<f", lambda x: round(x, 3)),
        ("delta_voltage", 209, "<f", lambda x: round(x, 3)),
        ("problem_code", 216, "B", lambda x: x if x in {1, 3, 7, 8, 9, 10, 11} else 0),
        ("balancer", 216, "B", lambda x: (x == 0x5)),
        ("balance_current", 217, "<f", lambda x: round(x, 3)),
    )

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: bytes = b""
        self._bms_info: dict[str, str] = {}
        self._exp_len: int = BMS._MIN_FRAME
        self._valid_reply: int = 0x02

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": pattern,
                "service_uuid": normalize_uuid_str("fee7"),
                "connectable": True,
            }
            for pattern in ("EK-*", "GW-*")
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ffe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe1"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        self._valid_reply = 0x01
        await self._await_msg(self._cmd(b"\x01"))
        self._valid_reply = 0x02
        return {
            "model": b2str(self._msg[8:24]),
            "hw_version": b2str(self._msg[24:32]),
            "sw_version": b2str(self._msg[32:40]),
        }

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Retrieve BMS data update."""

        if (
            len(self._frame) >= self._exp_len
            or not self._frame.startswith(BMS._HEAD_RSP)
        ) and data.startswith(BMS._HEAD_RSP):
            self._frame.clear()
            self._exp_len = max(
                int.from_bytes(data[6:8], byteorder="little", signed=False),
                BMS._MIN_FRAME,
            )

        self._frame += data

        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        # verify that data is long enough
        if len(self._frame) < self._exp_len:
            return

        if not self._frame.startswith(BMS._HEAD_RSP):
            self._log.debug("incorrect frame start.")
            return

        # trim message in case oversized
        if len(self._frame) > self._exp_len:
            self._log.debug("wrong data length (%i): %s", len(self._frame), self._frame)
            del self._frame[self._exp_len :]

        if self._frame[-1] != BMS._TAIL:
            self._log.debug("incorrect frame end.")
            return

        # check that message type is expected
        if self._frame[BMS._TYPE_POS] != self._valid_reply:
            self._log.debug(
                "unexpected message type 0x%X (length %i): %s",
                self._frame[BMS._TYPE_POS],
                len(self._frame),
                self._frame,
            )
            return

        if (crc := crc_sum(self._frame[:-2])) != self._frame[-2]:
            self._log.debug("invalid checksum 0x%X != 0x%X", self._frame[-2], crc)
            return

        self._msg = bytes(self._frame)
        self._msg_event.set()

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics and protocol state."""
        await super()._init_connection(char_notify)
        await self._fetch_device_info()
        self._valid_reply = 0x02  # cell information

    @staticmethod
    @cache
    def _cmd(cmd: bytes, reg: int = 0, value: list[int] | None = None) -> bytes:
        """Assemble a Neey BMS command."""
        value = [] if value is None else value
        assert len(value) <= 11
        frame: bytearray = bytearray(  # 0x14 frame length
            [*BMS._HEAD_CMD, cmd[0], reg & 0xFF, 0x14, *value]
        ) + bytes(11 - len(value))
        frame += bytes([crc_sum(frame), BMS._TAIL])
        return bytes(frame)

    @staticmethod
    def _cell_voltages(
        data: bytes,
        *,
        cells: int,
        start: int,
        size: int = 2,
        gap: int = 0,
        byteorder: Literal["little", "big"] = "little",
        divider: int = 1,
    ) -> list[float]:
        """Parse cell voltages from message."""
        return [
            round(value, 3)
            for idx in range(cells)
            if (value := unpack_from("<f", data, start + idx * size)[0])
        ]

    @staticmethod
    def _temp_sensors(data: bytes, sensors: int) -> list[int | float]:
        return [
            round(unpack_from("<f", data, 221 + idx * 4)[0], 2)
            for idx in range(sensors)
        ]

    @staticmethod
    def _conv_data(data: bytes) -> BMSSample:
        """Return BMS data from status message."""
        result: BMSSample = {}
        for key, idx, fmt, func in BMS._FIELDS:
            result[key] = func(unpack_from(fmt, data, idx)[0])

        return result

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        if not self._msg_event.is_set() or self._msg[4] != 0x02:
            # request cell info (only if data is not constantly published)
            self._log.debug("requesting cell info")
            await self._await_msg(data=BMS._cmd(b"\x02"))

        data: BMSSample = self._conv_data(self._msg)
        data["temp_values"] = BMS._temp_sensors(self._msg, 2)

        data["cell_voltages"] = BMS._cell_voltages(
            self._msg, cells=24, start=9, byteorder="little", size=4
        )

        self._msg_event.clear()  # clear event for next update
        return data
